# copyright (C) 1997-2000 Jean-Luc Fontaine (mailto:jfontain@free.fr)
# this program is free software: please read the COPYRIGHT file enclosed in this package or use the Help Copyright menu

set rcsId {$Id: snmp.tcl,v 1.8 2000/02/26 21:54:08 jfontain Exp $}


package provide snmp 1.0
package require Tnm 2.1.10

namespace eval snmp {

    variable nextRow 0

    array set data {
        updates 0
        switches {
            -a 1 --address 1 --community 1 --context 1 --delay 1 -i 1 --identifiers 1 --mibs 1 --port 1 --password 1 --retries 1
            -t 1 --table 1 --timeout 1 --trace 0 --trim 1 --user 1 --version 1 --window 1
        }
        pollTimes {10 5 20 30 60 120 300}
        sort {0 increasing}
    }
    set file [open [file join $::packageDirectory(snmp) snmp.htm]]
    set data(helpText) [read $file]                                                           ;# initialize HTML help data from file
    close $file

    proc printErrorAndExit {identifier message} {
        puts -nonewline stderr {snmp module:}
        if {[string length $identifier]>0} {
            puts -nonewline stderr " (mib: [mib file $identifier])"
        }
        puts stderr " $message"
        exit 1
    }

    proc initialize {optionsName} {
        upvar $optionsName options
        variable session
        variable trace

        if {![info exists options(-t)]&&![info exists options(--table)]} {
            printErrorAndExit {} {table must be specified}
        }
        catch {set trace $options(--trace)}
        if {[info exists options(--mibs)]} {                                               ;# comma separated list of MIB file names
            foreach file [split $options(--mibs) ,] {
                if {[catch {mib load $file} message]} {
                    printErrorAndExit {} $message
                }
            }
        }
        catch {set table $options(-t)}
        catch {set table $options(--table)}                                                                     ;# favor long option
        set string {}
        catch {set string $options(-i)}
        catch {set string $options(--identifiers)}                                                              ;# favor long option
        if {[info exists options(--trim)]} {
            processIdentifiers $table $string $options(--trim)
        } else {
            processIdentifiers $table $string
        }
        set arguments {}
        catch {set arguments [list -address $options(-a)]}
        catch {set arguments [list -address $options(--address)]}                                               ;# favor long option
        foreach switch {community context delay port password retries timeout window} {
            catch "lappend arguments -$switch $options(--$switch)"
        }
        if {[info exists options(--version)]} {                                                     ;# default is 1 (SNMP version 1)
            switch $options(--version) {
                1 {}
                2C {lappend arguments -version SNMPv2C}
                2U {lappend arguments -version SNMPv2U}
                default {
                    printErrorAndExit {} {version must be 1, 2C or 2U}
                }
            }
        }
        set session [eval snmp session $arguments]
    }

    proc validateIdentifier {name} {                                    ;# report an error if identifier is not part of a loaded MIB
        if {[catch {mib -exact syntax $name} message]} {
            printErrorAndExit {} $message
        }
    }

    proc processIdentifiers {table string {trim {}}} {
        # string format is identifier,identifier,...,identifier,,identifier,identifier,... with , to separate columns and ,, tables
        variable data
        variable requestIdentifiers
        variable requestLength
        variable indexLength
        variable indexColumns
        variable firstIdentifier
        variable numberOfColumns
        variable delta

        validateIdentifier $table
        set indices [mib -exact index $table]
        set index 0                                ;# initialize a few things since these identifiers are always part of the request
        foreach identifier $indices {
            set indexColumns($index) {}
            identifierType $identifier delta($index)
            incr index
        }
        set indexLength $index
        set entry [mib -exact successor $table]
        if {[llength $entry]>1} {
            printErrorAndExit $table "$table has several successors: please report case to jfontain@free.fr"
        }
        if {[string length $string]==0} {                                                                     ;# display all columns
            set identifiers [mib -exact index $string]                          ;# include indices which may not belong to the table
            foreach identifier [mib -exact successor $entry] {                                ;# columns are the successors of entry
                if {[lsearch -exact $identifiers $identifier]>=0} continue                            ;# already in identifiers list
                lappend identifiers $identifier
            }
            set views [list $identifiers]
        } else {                                                ;# generate list of views, a view being a list of column identifiers
            regsub -all ,, $string | string
            foreach list [split $string |] {
                lappend views [split $list ,]
            }
            foreach identifier [eval concat $views] {                                           ;# must be columns of the same table
                validateIdentifier $identifier
                if {[string compare $entry [mib -exact parent $identifier]]&&([lsearch -exact $indices $identifier]<0)} {
                    printErrorAndExit $identifier "$identifier is not a column of table $table"
                }
            }
        }
        set requestIdentifiers $indices                                                    ;# required to generate unique row number
        set firstIdentifier [mib oid [lindex $requestIdentifiers 0]]
        if {$indexLength>1} {                                                                           ;# multiple column index, so
            set data(0,label) {}      ;# include a row number column so that, for example, different views can be identically sorted
            set data(0,type) integer
            set data(0,message) {row creation order}
            set nextColumn 1
        } else {                                             ;# single column index: the user should have included it in his view(s)
            set nextColumn 0
        }
        foreach list $views {
            if {$indexLength>1} {
                set columns 0                                                      ;# creation order column is always included first
            } else {
                set columns {}
            }
            foreach identifier $list {
                set index [lsearch -exact $requestIdentifiers $identifier]                      ;# index in request identifiers list
                if {$index<0} {                                                                                   ;# not yet in list
                    set index [llength $requestIdentifiers]
                    lappend requestIdentifiers $identifier
                }
                if {[catch {set column $columnIndex($index)}]} {              ;# identifiers thus columns may be duplicated in views
                    set column $nextColumn
                    set columnIndex($index) $column
                    incr nextColumn
                }
                lappend indexColumns($index) $column                                      ;# there can be identical columns in views
                regsub ^$trim $identifier {} data($column,label)            ;# eventually trim string from left side of column title
                set data($column,type) [identifierType $identifier delta($index)]
                set data($column,message) [mib description $identifier]
                lappend columns $column
            }
            lappend viewsColumns $columns
        }
        set numberOfColumns $column
        set requestLength [llength $requestIdentifiers]
        if {[llength $viewsColumns]>1} {                                                                  ;# more than a single view
            foreach columns $viewsColumns {
                lappend data(views) [list visibleColumns $columns sort {0 increasing}]
            }
        }
    }

    proc identifierType {name counterName} {                                                           ;# must be a valid identifier
        upvar $counterName counter

        set counter 0
        switch [mib -exact syntax $name] {
            {OCTET STRING} {
                return ascii
            }
            {OBJECT IDENTIFIER} - IpAddress - TimeTicks - INTEGER - Integer32 - Unsigned32 {
                return dictionary                                         ;# integers are usually enumerations, such as ifOperStatus
            }
            Gauge - Gauge32 {
                return real                  ;# instead of integer since Tcl cannot handle positive integers greater than 0x7FFFFFFF
            }
            Counter - Counter32 {
                set counter 1
                return integer
            }
            Counter64 {
                set counter 1
                if {0x80000000>0} {                                                                                ;# 64 bit machine
                    return integer
                } else {
                    return real              ;# instead of integer since Tcl cannot handle positive integers greater than 0x7FFFFFFF
                }
            }
            default {
                return dictionary
            }
        }
    }

    proc update {} {
        variable requestIdentifiers
        variable currentRow

        catch {unset currentRow}
        getBulk $requestIdentifiers
    }

    proc getBulk {identifiers} {
        variable session
        variable trace

        if {[info exists trace]} {
            puts ">>> request(get-bulk-request):[formattedIdentifiers $identifiers]"
        }
        $session getbulk 1 0 $identifiers {::snmp::processResponse %E [list %V]}
    }

    proc formattedIdentifiers {list} {
        set string {}
        foreach identifier $list {
            append string " [mib name $identifier]"
        }
        return $string
    }

    proc formattedObjects {list} {                                                                               ;# list of varbinds
        set string {}
        foreach object $list {
            foreach {identifier type value} $object {}
            append string " [mib name $identifier]($value)"
        }
        return $string
    }

    proc processResponse {status objects} {       ;# objects are a list of varbinds (see snmp manual page) belonging to the same row
        variable requestLength
        variable firstIdentifier
        variable indexLength
        variable indexRow
        variable nextRow
        variable currentRow
        variable indexColumns
        variable last
        variable data
        variable delta
        variable trace

        if {[info exists trace]} {
            puts "<<< response($status):[formattedObjects $objects]"
        }
        if {[string compare $status noError]||([llength $objects]!=$requestLength)} {   ;# number of responses differs from requests
            cleanup
            incr data(updates)
            return                                                                                                           ;# done
        }
        set key {}                                                                                            ;# first determine row
        set index 0
        foreach object $objects {                                                                         ;# first determine the key
            foreach {identifier type value} $object {}
            if {($index==0)&&[string first $firstIdentifier $identifier]} {                                   ;# end of table passed
                cleanup
                incr data(updates)
                return                                                                                                       ;# done
            }
            lappend key $value                                                       ;# concatenate values in a list to make the key
            if {[incr index]==$indexLength} break
        }
        if {[catch {set row $indexRow($key)}]} {                                                                          ;# new row
            set row [set indexRow($key) $nextRow]
            incr nextRow
            if {$indexLength>1} {
                set data($row,0) $nextRow                                  ;# do not forget creation order column, which starts at 1
            }
        }
        set currentRow($row) {}                                                            ;# keep track of current rows for cleanup
        set index 0
        set identifiers {}
        foreach object $objects {                                                                         ;# now fill row data cells
            foreach {identifier type value} $object {}
            lappend identifiers $identifier
            if {$delta($index)} {
                set current $value
                if {[info exists last($row,$index)]} {
                    set value [expr {$current-$last($row,$index)}]                                    ;# display deltas for counters
                } else {
                    set value ?                                                       ;# need at least 2 values to make a difference
                }
                set last($row,$index) $current
            }
            foreach column $indexColumns($index) {
                set data($row,$column) $value                                                ;# fill the columns for that identifier
            }
            incr index
        }
        getBulk $identifiers                                                                         ;# keep going till end of table
    }

    proc cleanup {} {                                                                 ;# remove disappeared rows and associated data
        variable currentRow
        variable indexRow
        variable data
        variable last
        variable numberOfColumns

        foreach {key row} [array get indexRow] {
            if {[info exists currentRow($row)]} continue
            unset indexRow($key)
            for {set column 0} {$column<$numberOfColumns} {incr column} {
                unset data($row,$column)
            }
            foreach name [array names last $row,*] {                                      ;# for Tcl 8.3, use new array unset syntax
                unset last($name)
            }
        }
        catch {unset currentRow}
    }

}
